/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.common.graph;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.tudarmstadt.maki.simonstrator.api.Graphs;

/**
 * A basic class realizing the Graph-Interface
 * 
 * @author Michael Stein
 * @author Roland Kluge
 * @author Bjoern Richerzhagen
 *
 */
public class BasicGraph implements Graph {

	private final Set<INode> nodes = new LinkedHashSet<>();

	private Map<INodeID, INode> nodesById = new LinkedHashMap<>();

	private final Set<IEdge> edges = new LinkedHashSet<>();

	// The list of all edges. For performance reasons
	private final Map<EdgeID, IEdge> edgesById = new LinkedHashMap<>();

	// Assigns to each node the set of outgoing edges of this node
	private final Map<INodeID, Set<IEdge>> outgoingAdjacencyList = new HashMap<>();

	// Assigns to each node the set of incoming edges of this node
	private final Map<INodeID, Set<IEdge>> incomingAdjacencyList = new HashMap<>();

	/**
	 * Creates an empty graph. Components should create new graphs by using the
	 * {@link Graphs} class in the API
	 */
	public BasicGraph() {
	}

	/**
	 * Creates a graph from the given nodes and edges.
	 * 
	 * Duplicate nodes and edges will be ignored
	 * 
	 * @param nodes
	 *            The set of nodes contained in the graph.
	 * @param edges
	 *            Set of edges contained in the graph. Assumes that every start
	 *            and end node in the edges is also contained in the nodes set
	 */
	public BasicGraph(Iterable<? extends INode> nodes,
			Iterable<? extends IEdge> edges) {
		this();

		for (final INode node : nodes)
			this.addNode(node);

		for (final IEdge edge : edges)
			this.addEdge(edge);
	}

	/**
	 * Creates an edge-induced graph.
	 * 
	 * The nodes are derived from the edges.
	 * 
	 * @param edges
	 *            the edges of the graph
	 */
	public BasicGraph(Iterable<? extends IEdge> edges) {
		this();
		for (final IEdge edge : edges) {
			this.addNode(this.createNode(edge.fromId()));
			this.addNode(this.createNode(edge.toId()));
		}
		this.addElements(edges);
	}

	public static BasicGraph createGraphFromNodeIdsAndEdges(
			Iterable<? extends INodeID> nodes,
			Iterable<? extends IEdge> edges) {
		final BasicGraph graph = new BasicGraph();
		for (INodeID nodeId : nodes) {
			graph.addNode(new Node(nodeId));
		}
		for (IEdge edge : edges) {
			graph.addEdge(edge);
		}
		return graph;
	}

	/**
	 * Creates an edge-induced graph.
	 * 
	 * The nodes are derived from the end nodes of the edges.
	 * 
	 * @param edges
	 *            the edges of the graph
	 */
	@SafeVarargs
	public BasicGraph(IEdge... edges) {
		this(Arrays.asList(edges));
	}

	@Override
	public IEdge createEdge(INodeID from, INodeID to) {
		IEdge edge = getEdge(from, to);
		return (edge == null ? new DirectedEdge(from, to) : edge);
	}

	@Override
	public IEdge createEdge(INodeID from, INodeID to, double weight) {
		IEdge edge = createEdge(from, to);
		if (edge == null) {
			return Graphs.createDirectedWeightedEdge(from, to, weight);
		} else {
			edge.setProperty(GenericGraphElementProperties.WEIGHT, weight);
			return edge;
		}
	}

	@Override
	public INode createNode(INodeID id) {
		if (this.containsNode(id)) {
			return this.getNode(id);
		} else {
			Node newNode = new Node(id);
			this.addNode(newNode);
			return newNode;
		}
	}

	@Override
	public boolean addElement(IElement element) {
		if (element instanceof INode) {
			return this.addNode((Node) element);
		} else if (element instanceof IEdge) {
			return this.addEdge((IEdge) element);
		} else {
			throw new AssertionError("Unknown graph element type.");
		}
	}

	@Override
	public boolean add(IElement element) {
		return this.addElement(element);
	}

	@Override
	public boolean addEdge(IEdge edge) {
		this.ensureThatNodeIsInGraph(edge.toId());
		this.ensureThatNodeIsInGraph(edge.fromId());
		if (!this.containsEdge(edge)) {
			this.edges.add(edge);
			this.edgesById.put(edge.getId(), edge);
			this.outgoingAdjacencyList.get(edge.fromId()).add(edge);
			this.incomingAdjacencyList.get(edge.toId()).add(edge);
				return true;
		} else {
			return false;
				}
				}

	@Override
	public boolean addNode(INode node) {
		if (!this.containsNode(node)) {
			this.nodes.add(node);
			this.nodesById.put(node.getId(), node);
			this.incomingAdjacencyList.put(node.getId(), new HashSet<IEdge>());
			this.outgoingAdjacencyList.put(node.getId(), new HashSet<IEdge>());
				return true;
		} else {
			return false;
			}
		}

	@Override
	public void createAndAddNode(INodeID nodeId) {
		this.addNode(this.createNode(nodeId));
	}

	@Override
	public void createAndAddNodes(Iterable<INodeID> nodeIds) {
		for (final INodeID nodeId : nodeIds)
			this.createAndAddNode(nodeId);
	}

	@Override
	public boolean removeElement(IElement element) {
		if (element instanceof INode) {
			return removeNode((INode) element);
		} else if (element instanceof IEdge) {
			return removeEdge((IEdge) element);
		} else {
			throw new AssertionError("Unknown graph element type.");
		}
					}

	@Override
	public boolean removeNode(INodeID node) {
		if (this.containsNode(node))
			return this.removeNode(this.getNode(node));
		return false;
				}

	private boolean removeNode(INode node) {
		if (!this.containsNode(node.getId()))
			return false;

		final List<IEdge> edgesToRemove = new ArrayList<>();
		for (final IEdge edge : this.getOutgoingEdges(node.getId()))
			edgesToRemove.add(edge);

		for (final IEdge edge : this.getIncomingEdges(node.getId()))
			edgesToRemove.add(edge);

		for (final IEdge edgeToRemove : edgesToRemove)
			this.removeEdge(edgeToRemove);

		nodesById.remove(node.getId());
		nodes.remove(node);

				return true;
			}

	@Override
	public boolean removeEdge(IEdge edge) {
		if (!this.containsEdge(edge))
			return false;

		this.incomingAdjacencyList.get(edge.toId()).remove(edge);
		this.outgoingAdjacencyList.get(edge.fromId()).remove(edge);
		this.edges.remove(edge);
		this.edgesById.remove(edge.getId());
		return true;
	}

	@Override
	public Set<INode> getNodes() {
		return nodes;
	}

	@Override
	public Iterable<INodeID> getNodeIds() {
		return this.nodesById.keySet();
	}

	@Override
	public Set<EdgeID> getEdgeIds() {
		return this.edgesById.keySet();
	}

	@Override
	public int getNodeCount() {
		return this.getNodes().size();
	}

	@Override
	public Set<IEdge> getEdges() {
		return Collections.unmodifiableSet(this.edges);
	}

	@Override
	public int getEdgeCount() {
		return this.getEdges().size();
	}

	@Override
	public boolean contains(IElement element) {
		if (element instanceof INode)
			return containsNode((INode) element);
		else if (element instanceof IEdge)
			return containsEdge((IEdge) element);
		else
			throw new IllegalArgumentException(
					"Unsupported kind of element: " + element);
	}

	@Override
	public boolean containsNode(INodeID nodeId) {
		return nodesById.containsKey(nodeId);
	}

	@Override
	public boolean containsNode(INode node) {
		return this.containsNode(node.getId());
	}

	@Override
	public INode getNode(INodeID nodeId) {
		return nodesById.get(nodeId);
	}

	@Override
	public IEdge getEdge(INodeID from, INodeID to) {
		for (IEdge edge : edges) {
			if (edge.toId().equals(to) && edge.fromId().equals(from)) {
				return edge;
			}
		}
		return null;
	}

	@Override
	public IEdge getEdge(EdgeID edgeID) {
		return this.edgesById.get(edgeID);
	}

	/**
	 * Computes subgraph of the currently known local view. This operation
	 * changes the current graph instead of cloning it.
	 */
	@Override
	public BasicGraph getLocalView(INodeID node, int k) {
		return getLocalView(node, k, false);
	}

	/**
	 * Computes subgraph of the currently known local view. The clone parameter
	 * specifies whether this Graph instance is changed. If clone is true, a new
	 * Graph instance is generated. Elsewise, this Graph instance is modified.
	 */
	@Override
	public BasicGraph getLocalView(INodeID center, int k, boolean clone) {
		BasicGraph g = clone ? this.clone() : this;

		// filter nodes
		Set<INodeID> kHopNodes = getNeighbors(center, k, false);

		// determine edges
		Set<IEdge> kHopEdges = new LinkedHashSet<>();
		for (final INodeID kHopNode : kHopNodes) {
			for (final IEdge outgoingEdge : this.getOutgoingEdges(kHopNode)) {
				if (kHopNodes.contains(outgoingEdge.toId())) {
					kHopEdges.add(outgoingEdge);
				}
			}
		}

		List<INode> nodesToBeRemoved = new ArrayList<>();
		for (final INode node : this.getNodes()) {
			if (!kHopNodes.contains(node.getId())) {
				nodesToBeRemoved.add(node);
			}
			}
		for (final INode node : nodesToBeRemoved) {
			g.removeElement(node);
		}

		assert checkGraphConsistency();

		return g;
	}

	@Override
	public void clear() {
		this.nodes.clear();
		this.edgesById.clear();
		this.edges.clear();
		this.incomingAdjacencyList.clear();
		this.outgoingAdjacencyList.clear();
	}

	/**
	 * Returns the nodes contained in the k-hop neighborhood of the specified
	 * node. This includes the node itself.
	 * 
	 * If the directedNeighborhood flag is true, edge directions are considered
	 * for the neighborhood
	 */
	@Override
	public Set<INodeID> getNeighbors(INodeID center, int k,
			boolean directedNeighborhood) {
		return getNeighbors(center, k, directedNeighborhood,
				new LinkedHashSet<INodeID>());
	}

	/**
	 * Checks that every start and end node in the edges is also contained in
	 * the nodes set
	 */
	private boolean checkGraphConsistency() {
		for (IEdge directedEdge : edges) {
			if (!nodesById.containsKey(directedEdge.fromId())
					|| !nodesById.containsKey(directedEdge.toId())) {
				return false;
			}
		}

		return true;
	}

	// helper method
	private Set<INodeID> getNeighbors(INodeID node, int k,
			boolean directedNeighborhood, LinkedHashSet<INodeID> visited) {

		if (k < 0)
			throw new IllegalArgumentException("k must be >=0, but was " + k);

		if (k == 0) {
			visited.add(node);
			LinkedHashSet<INodeID> oneNodeNeighborSet = new LinkedHashSet<INodeID>();
			oneNodeNeighborSet.add(node);
			return oneNodeNeighborSet;
		}

		visited.add(node);

		LinkedHashSet<INodeID> neighbors = new LinkedHashSet<INodeID>();
		neighbors.add(node);

		for (INodeID neighbor : getNeighbors(node, directedNeighborhood)) {
			if (!visited.contains(neighbor))
				neighbors.addAll(getNeighbors(neighbor, k - 1,
						directedNeighborhood, visited));
		}

		return neighbors;
	}

	@Override
	public Set<IEdge> getOutgoingEdges(INodeID node) {
		ensureThatNodeIsInGraph(node);
		return Collections
				.unmodifiableSet(this.outgoingAdjacencyList.get(node));
			}

	@Override
	public Set<IEdge> getIncomingEdges(INodeID node) {
		ensureThatNodeIsInGraph(node);
		return Collections
				.unmodifiableSet(this.incomingAdjacencyList.get(node));
			}

	@Override
	public Set<INodeID> getNeighbors(INodeID node,
			boolean directedNeighborhood) {
		Set<INodeID> neighbors = new LinkedHashSet<INodeID>();
		for (final IEdge edge : this.getOutgoingEdges(node))
				neighbors.add(edge.toId());

		// Also add start nodes of incoming links
		if (!directedNeighborhood)
			for (final IEdge edge : this.getIncomingEdges(node))
				neighbors.add(edge.fromId());

		return Collections.unmodifiableSet(neighbors);
			}

	@Override
	public Set<INodeID> getNeighbors(INodeID node) {
		return getNeighbors(node, true);
	}

	@Override
	public Set<INodeID> getNeighbors(INode node) {
		return getNeighbors(node.getId());
	}

	private void ensureThatNodeIsInGraph(INodeID node) {
		if (!this.nodesById.containsKey(node))
			throw new IllegalStateException(
					"Node " + node + " is not contained in the graph");
	}

	@Override
	public BasicGraph clone() {
		/*
		 * As per the API contract, the inner objects are NOT cloned. This
		 * cloned graph is just a new view.
		 */
		BasicGraph g = new BasicGraph();
		g.addNodes(this.nodes);
		g.addEdges(this.edges);
		return g;
	}

	@Override
	public void addEdges(Iterable<? extends IEdge> edges) {
		for (IEdge edge : edges) {
			this.addEdge(edge);
		}
		}

	@Override
	public void addNodes(Iterable<? extends INode> nodes) {
		for (INode node : nodes) {
			this.addNode(node);
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("Graph [V= ");
		sb.append(this.nodes);
		sb.append(", E= ");
		sb.append(this.edgesById);
		sb.append("]");
		return sb.toString();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((edges == null) ? 0 : edges.hashCode());
		result = prime * result + ((nodes == null) ? 0 : nodes.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		BasicGraph other = (BasicGraph) obj;
		if (edges == null) {
			if (other.edges != null)
				return false;
		} else if (!edges.equals(other.edges))
			return false;
		if (nodes == null) {
			if (other.nodes != null)
				return false;
		} else if (!nodes.equals(other.nodes))
			return false;
		return true;
	}

	@Override
	public void addElements(Iterable<? extends IElement> elements) {
		for (final IElement element : elements) {
			if (element instanceof INode)
				this.addNode((INode) element);
		}
		for (final IElement element : elements) {
			if (element instanceof IEdge)
				this.addEdge((IEdge) element);
		}
	}

	@Override
	public boolean containsEdge(IEdge edge) {
		return this.edges.contains(edge);
	}

	@Override
	public int getOutdegree(INodeID id) {
		return this.getOutgoingEdges(id).size();
	}

	@Override
	public int getIndegree(INodeID node) {
		return this.getIncomingEdges(node).size();
	}

	@Override
	public Set<INodeID> getNeighbors(INode node, int k,
			boolean directedNeighborhood) {
		return getNeighbors(node.getId(), k, directedNeighborhood);
	}

	@Override
	public Set<INodeID> getNeighbors(INode node, boolean directedNeighborhood) {
		return getNeighbors(node.getId(), directedNeighborhood);
	}

	@Override
	public boolean equalsAttributes(Graph otherGraph) {
		if (otherGraph.getNodeCount() != this.getNodeCount()
				|| otherGraph.getEdgeCount() != this.getEdgeCount())
			return false;
		for (final INode node : this.getNodes()) {
			if (!otherGraph.containsNode(node))
				return false;
		}
		for (final IEdge edge : this.getEdges()) {
			if (!otherGraph.containsEdge(edge))
				return false;
		}
		return false;
	}

	@Override
	public boolean isBidirectional() {
		if (this.getEdgeCount() % 2 != 0)
			return false;

		Set<IEdge> reverseEdgeSet = new HashSet<>();
		for (final IEdge edge : this.getEdges()) {
			reverseEdgeSet.add(new DirectedEdge(edge.toId(), edge.fromId()));
		}
		for (final IEdge edge : this.getEdges()) {
			if (!this
					.containsEdge(new DirectedEdge(edge.fromId(), edge.toId())))
				return false;

		}
		return true;
	}

	@Override
	public Iterable<INodeID> getIsolatedNodes() {
		final List<INodeID> isolatedNodes = new ArrayList<>();
		for (final INode node : this.getNodes()) {
			if (this.getOutdegree(node.getId()) == 0
					&& this.getIndegree(node.getId()) == 0) {
				isolatedNodes.add(node.getId());
			}
		}
		return isolatedNodes;
	}

	@Override
	public Iterable<INodeID> getRootNodes() {
		final List<INodeID> rootNodes = new ArrayList<>();
		for (final INode node : this.getNodes()) {
			if (this.getIndegree(node.getId()) == 0) {
				rootNodes.add(node.getId());
			}
		}
		return rootNodes;
	}

	@Override
	public Iterable<INodeID> getLeafNodes() {
		final List<INodeID> leafNodes = new ArrayList<>();
		for (final INode node : this.getNodes()) {
			if (this.getOutdegree(node.getId()) == 0) {
				leafNodes.add(node.getId());
			}
		}
		return leafNodes;
	}

	@Override
	public Set<IEdge> getOutgoingEdges(INode node) {
		return this.getOutgoingEdges(node.getId());
	}

	@Override
	public Set<IEdge> getIncomingEdges(INode node) {
		return this.getIncomingEdges(node.getId());
	}

	@Override
	public boolean remove(IElement element) {
		return this.removeElement(element);
	}
}
